part of 'development_by_muscle_bloc.dart';

@immutable
abstract class DevelopmentByMuscleState extends Equatable {
  const DevelopmentByMuscleState();

  @override
  List<Object> get props => [];
}

class DevelopmentByMuscleStateInitial extends DevelopmentByMuscleState {
  const DevelopmentByMuscleStateInitial();
}

class DevelopmentByMuscleLoadingState extends DevelopmentByMuscleState {
  const DevelopmentByMuscleLoadingState();
}

class DevelopmentByMuscleReadyState extends DevelopmentByMuscleState {
  const DevelopmentByMuscleReadyState();
}

class DevelopmentByMuscleErrorState extends DevelopmentByMuscleState {
  final String message;
  const DevelopmentByMuscleErrorState({required this.message});

  @override
  List<Object> get props => [message];
}
